import logging

from movierender.render.pipelines import PipelineException


class ImagePipeline:
    def __init__(self, *args, ax=None, zstack=0, **kwargs):
        self._kwargs = kwargs
        self.ax = ax
        self.zstack = zstack
        self.logger = logging.getLogger(__name__)

        # if len(args) > 0 and isinstance(args[0], MovieRenderer):
        if len(args) > 0 and args[0].__class__.__name__[-13:] == 'MovieRenderer':
            self._renderer = args[0]
        else:
            self._renderer = None

    def __radd__(self, ovrl):
        # if isinstance(ovrl, MovieRenderer):
        if ovrl.__class__.__name__[-13:] == 'MovieRenderer':
            ovrl_pipeline_filled = len(ovrl.image_pipeline) > 0
            ovrl_pipeline_empty = not ovrl_pipeline_filled
            at_least_one_ax_in_pipeline = ovrl_pipeline_filled and all([ip.ax is None for ip in ovrl.image_pipeline])
            if (ovrl_pipeline_empty and at_least_one_ax_in_pipeline) \
                    or (ovrl_pipeline_filled and self.ax is None):
                raise PipelineException(
                    f"More than one image processing pipeline when adding {ovrl.__class__.__name__}. "
                    "If you need to add more image pipelines, "
                    "consider providing an ax parameter to the class constructor.")
            else:
                assert len(ovrl.image.frames) > 0, "No images to process."
                ovrl.image_pipeline.append(self)
                self._renderer = ovrl
                return ovrl
        return self

    def __call__(self, *args, **kwargs):
        raise NotImplementedError



from pathlib import Path

import typer
from fileops.export.config import search_config_files
from fileops.logger import get_logger
from typing_extensions import Annotated

from movierender.scripts.render import render_configuration_file_cmd

log = get_logger(name='render-folder')


def render_folder_cmd(
        path: Annotated[
            Path, typer.Argument(help="Path where configuration files are located. "
                                      "If no path is given, the current folder will be used.")] = None,
):
    if path is None:
        log.info(f"No path provided")
        path = Path('.').absolute()
    cfg_path_list = search_config_files(path)

    total_rendered = 0
    for cfg_path in cfg_path_list:
        if cfg_path.parent.name[0:3] == "bad":
            continue

        log.info(f"Reading configuration file {cfg_path}")
        render_configuration_file_cmd(cfg_path)
        total_rendered += 1

    return total_rendered



from fileops.logger import get_logger, silence_loggers
from typer import Typer

from ._render_configfile import render_configuration_file_cmd
from ._render_folder import render_folder_cmd
from ._render_movie import render_movie_cmd
from ._render_panel import render_panel_cmd

log = get_logger(name='render', debug=False)
silence_loggers(loggers=["tifffile", "matplotlib", "PIL"], output_log_file="silenced.log")
app = Typer()

app.command(name='file')(render_configuration_file_cmd)
app.command(name='folder')(render_folder_cmd)
app.command(name='movie')(render_movie_cmd)
app.command(name='panel')(render_panel_cmd)

if __name__ == "__main__":
    app()



from __future__ import annotations

import logging
import os
import uuid
from pathlib import Path
from typing import List, TYPE_CHECKING

import imageio.v3 as iio
import moviepy.editor as mpy
import numpy as np
import skimage
from fileops.export.config import ConfigMovie
from fileops.image import ImageFile
from fileops.image.exceptions import FrameNotFoundError
from fileops.pathutils import ensure_dir
from matplotlib.figure import Figure

from movierender.render.pipelines import SingleImage, ImagePipeline

if TYPE_CHECKING:
    from movierender.overlays import Overlay


class SequentialMovieRenderer:
    layers: List[Overlay]
    image: ImageFile

    def __init__(self, fig: Figure, config: ConfigMovie, show_axis=False, invert_y=False, **kwargs):
        self._kwargs = {
            'fontdict': {'size': 10},
        }
        self._kwargs.update(**kwargs)

        self.fig = fig
        self.show_axis = show_axis
        self.ax = fig.gca()

        self.layers = []
        self.logger = logging.getLogger(__name__)

        self.time = 0
        self.frame = 0
        self.fps = config.fps
        self.duration = None
        self.bitrate = config.bitrate

        imf = config.image_file
        self._cfg = config
        self.image_pipeline: List[ImagePipeline] = []
        self.image = imf
        self.inv_y = invert_y
        self._last_f = imf.frames[-1]
        self._max_frame = max(self._cfg.frames)
        self._frame_offset = min(self._cfg.frames)  # used when frames start at a number greater than zero
        self._render = np.zeros((imf.width, imf.height), dtype=float)
        self._load_image()

        self._tmp = Path(os.curdir) / 'tmp' / 'render' / Path(imf.base_path).name / str(uuid.uuid4())
        ensure_dir(self._tmp)

    def __iter__(self):
        return self

    def __next__(self):
        if self.image_pipeline is not None:
            imp = self.image_pipeline
        else:
            imp = SingleImage(self)
        self.time = (self.time + 1) % self._max_frame

        return imp(invert_y=self.inv_y)

    def __getattr__(self, name):
        if name in self._kwargs:
            return self._kwargs[name]
        else:
            raise AttributeError("No such attribute: " + name)

    def _load_image(self):
        assert len(self.image.frames) > 1, "More than one frame needed to make a movie."
        self.logger.info(f"Loaded {self.image.image_path}. WxH({self.image.width},{self.image.height}), "
                         f"channels: {len(self.image.channels)}, "
                         f"frames: {len(self.image.frames)}, stacks: {len(self.image.zstacks)}")
        self.duration = len(self.image.frames) / self.fps
        self.logger.info(f"Total duration {self.duration:.3f}[s]")

        self._kwargs.update({
            'pix_per_um': self.image.pix_per_um,
            'timestamps': self.image.timestamps,
            'dt':         self.image.time_interval,
            'n_frames':   len(self.image.frames),
            'n_channels': len(self.image.channels),
            'n_zstacks':  len(self.image.zstacks),
            'width':      self.image.width,
            'height':     self.image.height,
            # 'ranges':     self.image.intensity_ranges
        })

    def render(self, filename=None, test=False):
        """
        Render the movie into an mp4 file.
        """

        def make_frame_mpl(t):
            self.time = t
            # calculate frame given time
            _fr_from_t = int(round(self.fps * t)) + self._frame_offset
            self.frame = min(_fr_from_t, self._max_frame)

            if self.frame == self._last_f:
                return self._render

            try:
                self._render = iio.imread(self._tmp.joinpath(f"f{self.frame:05d}.png"))
                self._last_f = self.frame

                # row, col, ch = self._render.shape
                # if ch == 3:
                #     self._render = self._render[:, :, 0:3]
                self._render = self._render[:, :, 0:3]
            except FileNotFoundError:
                self.logger.warning(f"frame {self.frame} not rendered, so using last rendered one")

            # self.logger.debug(f"loaded image of shape {self._render.shape}")
            return self._render

        def render_frame(frame):

            self.logger.info(f"rendering frame {frame}")
            self.frame = frame
            # calculate time given frame
            self.time = (frame - self._frame_offset) / self.fps

            img_path = self._tmp.joinpath(f"f{frame:05d}.png")
            if os.path.exists(img_path):
                self.logger.warning(f'File {img_path.name} already exists in folder {img_path.parent.name}.')
                return

            # clear axes of all objects
            self.ax.cla()
            for ovrl in self.layers:
                if ovrl.ax is not None:
                    ovrl.ax.cla()
            for imgp in self.image_pipeline:
                if imgp.ax is not None:
                    imgp.ax.cla()
                if not self.show_axis and imgp.ax is not None:
                    imgp.ax.set_xticklabels([])
                    imgp.ax.set_yticklabels([])
                    imgp.ax.set_xticks([])
                    imgp.ax.set_yticks([])

            for imgp in self.image_pipeline:
                ppu = self.image.pix_per_um if self.image.pix_per_um is not None else 1
                ext = (0, self.image.width / ppu, 0, self.image.height / ppu)
                ax = imgp.ax if imgp.ax is not None else self.ax
                img = imgp()
                img = skimage.util.img_as_float(img)
                ax.imshow(img, cmap='gray', extent=ext,
                          origin='upper' if self.inv_y else 'lower',
                          interpolation='none', aspect='equal',
                          zorder=0)
                for ovrl in self.layers:
                    kwargs = self._kwargs.copy()
                    kwargs.update(**ovrl._kwargs, show_axis=self.show_axis)
                    ovrl.plot(ax=self.ax if ovrl.ax is None else None, **kwargs)

            for ovrl in self.layers:
                if not ovrl.show_axis and ovrl.ax is not None:
                    ovrl.ax.set_xticklabels([])
                    ovrl.ax.set_yticklabels([])
                    ovrl.ax.set_xticks([])
                    ovrl.ax.set_yticks([])
            self.fig.tight_layout()
            self.fig.savefig(img_path, facecolor='white', transparent=False)

        # Start of method
        if filename is None:
            _, filename = os.path.split(self._file)
            filename += ".mp4"
        rendered_frames = list()
        for fr in sorted(self._cfg.frames):
            try:
                render_frame(fr)
                rendered_frames.append(fr)
            except FrameNotFoundError:
                continue

        dur = len(rendered_frames) / self.fps
        animation = mpy.VideoClip(make_frame_mpl, duration=dur)
        animation.write_videofile(filename,
                                  fps=self._cfg.fps,
                                  bitrate=self._cfg.bitrate,
                                  # codec='libx264',
                                  # audio_codec='pcm_s32le',
                                  ffmpeg_params=[
                                      '-vf', 'pad=ceil(iw/2)*2:ceil(ih/2)*2',
                                      '-pix_fmt', 'yuv420p'
                                  ])
        animation.close()

    def __repr__(self):
        return f"<MovieRender object (sequential) at {hex(id(self))}> with {len(self._kwargs)} arguments."



from typing import List

from matplotlib import patches
from roifile import ImagejRoi

from .overlay import Overlay


class ImagejROI(Overlay):
    def __init__(self, roi_list: List[ImagejRoi] = None, **kwargs):
        assert all(r is not None for r in roi_list), "Need ROIs to render on axes."
        self.roi_lst = roi_list
        super().__init__(**kwargs)

    def plot(self, ax=None, xy=(0, 0), fontdict=None, **kwargs):
        if ax is None:
            ax = self.ax
        assert ax is not None, "No axes found to plot overlay."

        # scale = self._renderer.image.pix_per_um
        scale = self._renderer.image.um_per_pix
        for roi in self.roi_lst:
            w = abs(roi.right - roi.left) * scale
            h = abs(roi.top - roi.bottom) * scale
            r, c = (self._renderer.image.height - roi.bottom) * scale, roi.left * scale
            # ax.text(c + 0.5 * w, r + 1.2 * h, s, ha="center", va="center", c="w", size=5)

            # color = self._style[s]['color'] if self._style is not None and 'color' in self._style[s] else None
            # color = color if color is not None else 'w'
            color = 'yellow'

            rect = patches.Rectangle((c, r), w, h, linewidth=1, edgecolor=color, facecolor='none', zorder=100)
            ax.add_patch(rect)



from .overlay import Overlay


class ScaleBar(Overlay):
    def plot(self, ax=None, um=None, xy=None, lw=None, fontdict=None, **kwargs):
        if ax is None:
            ax = self.ax
        assert ax is not None, "No axes found to plot overlay."

        lw = lw if lw is not None else self._kwargs.get("lw", 1)
        um = um if um is not None else self._kwargs.get("um", None)
        x0, y0 = xy if xy is not None else self._kwargs.get("xy", (0, 0))

        if um is None:
            return

        ax.plot([x0, x0 + um], [y0, y0], c='w', lw=lw, zorder=1000)
        if kwargs.get("show_text", True):
            # Add text below the scalebar
            sbar_lw2_um = 352.77778 * lw / 100 / 2
            ax.text(x0 + um / 2, y0 + sbar_lw2_um, f'{um} um', color='w', fontdict=fontdict,
                    horizontalalignment='center', zorder=1000)


def secs_to_string(secs: int, string_format="hh:mm:ss"):
    mins = int(secs / 60)
    hours = int(mins / 60)
    mins -= hours * 60
    secs -= hours * 3600 + mins * 60

    if string_format == "hh:mm:ss":
        txt = f'{hours:02d}:{mins:02d}:{int(secs):02d}'
    elif string_format == "hh:mm":
        txt = f'{hours:02d}:{mins:02d}'
    elif string_format == "mm:ss":
        if hours < 0:
            txt = f'{mins:02d}:{secs:02d}'
        else:
            txt = f'{hours:02d}:{mins:02d}:{secs:02d}'
    else:
        raise Exception("Timestamp string format not implemented.")

    return txt


class Timestamp(Overlay):
    def plot(self, ax=None, xy=(0, 0), string_format="hh:mm:ss", timestamps=None, fontdict=None, va=None, color='white',
             **kwargs):
        if ax is None:
            ax = self.ax
        assert ax is not None, "No axes found to plot overlay."
        assert timestamps is not None, "Need timestamps to render on axes."
        if va is None:
            va = 'center'

        x0, y0 = xy
        txt = ''
        if timestamps:
            _secs0 = timestamps[self._renderer.frame - 1] if len(timestamps) >= self._renderer.frame else timestamps[-1]
            _secs1 = (self._renderer.frame - 1) * self._renderer.image.time_interval

            secs = int(max(_secs0, _secs1))
            txt = secs_to_string(secs, string_format)

        txt = f'{self._renderer.frame}  {txt}'
        ax.text(x0, y0, txt, color=color, fontdict=fontdict, verticalalignment=va, zorder=1000)



import matplotlib.pyplot as plt

from .overlay import Overlay


class Treatment(Overlay):
    def plot(self, ax=None, xy=(0, 0), lw=1, fontdict=None, expdict=None, **kwargs):
        if ax is None:
            ax = self.ax
        assert ax is not None, "No axes found to plot overlay."
        assert expdict is not None, "Experiment label parameters needed to apply this overlay."

        x0, y0 = xy
        r = self._renderer.fig.canvas.get_renderer()

        # get dimensions of the dot
        dot = plt.scatter(x0, y0, s=fontdict['size'] ** 2, c=None, lw=0)
        bb = dot.get_paths()[0].get_extents(transform=ax.transData)

        for name, settings in expdict.items():
            if self._renderer.frame in settings['on']:
                ax.scatter(x0, y0, s=fontdict['size'] ** 2, c=settings['color'], lw=lw)
            else:
                ax.scatter(x0, y0, s=fontdict['size'] ** 2, facecolor='black', edgecolor='w', lw=lw, hatch='//////')

            _x = 10 * bb.width / self._renderer.pix_per_um
            plt.annotate(f'{name}',  # this is the text
                         (x0, y0),  # this is the point to label
                         textcoords="offset points",  # how to position the text
                         xytext=(_x, 0),  # distance from text to points (x,y)
                         ha='left',  # horizontal alignment can be left, right or center
                         va='center_baseline',  # vertical can be center, top, bottom, baseline, center_baseline
                         c='white')

            y0 -= 1.5 * bb.height / self._renderer.pix_per_um



import datetime
import time as tim

import numpy as np
import pandas as pd
from matplotlib import dates

from .overlay import Overlay


class DataTimeseries(Overlay):
    def __init__(self, df: pd.DataFrame, x="time", y=None, frame="frame", time="time", style_dict=None, **kwargs):
        assert all([it in df.columns for it in [x, y]]), "Data point columns not found in DataFrame."
        self._x = x
        self._y = y
        self._f = frame
        self._t = time
        self._style = style_dict
        # Rename columns if parameters were given
        self.df = (df
                   .rename(columns={frame: 'frame', time: 'time'})
                   .assign(x=df[x], y=df[y]))

        # assuming 'x' values are time in seconds
        # format seconds to date with this format '%H:%M:%S'
        self.df.loc[:, 'x'] = list(
            map(datetime.datetime.strptime, map(lambda s: tim.strftime('%H:%M:%S', tim.gmtime(s)), self.df['x']),
                len(self.df['x']) * ['%H:%M:%S']))
        self.df.loc[:, 'x'] = pd.to_datetime(self.df.loc[:, 'x'])

        super().__init__(**kwargs)

    def plot(self, ax=None, legend=False, plot_dots=True, lw=2, **kwargs):
        if ax is None:
            ax = self.ax
        assert ax is not None, "No axes found to plot overlay."
        # assert timestamps is not None, "Need timestamps to render on axis."

        xmin, xmax = self.df['x'].min(), self.df['x'].max()
        ymin, ymax = self.df['y'].min(), self.df['y'].max()
        if not np.isfinite([ymin, ymax]).all():
            return

        fr = self._renderer.frame if self._renderer is not None else self.df['frame'].max()
        dat = self.df.query("frame <= @fr")

        if plot_dots:
            ax.scatter(dat['x'].tolist(), dat['y'], s=1, c='gray', zorder=10)
        for (ix_u, ix_sig), uniseries in self.df.groupby(['unit', 'signal']):
            ax.plot(uniseries['x'], uniseries['y'], c='gray', lw=0.5, zorder=10)
        for (ix_u, ix_sig), uniseries in dat.groupby(['unit', 'signal']):
            color = self._style[ix_sig]['color'] if self._style is not None and 'color' in self._style[ix_sig] else None
            color = color if color is not None else 'r'
            ax.plot(uniseries['x'], uniseries['y'], c=color, lw=lw, zorder=20)

        ax.set_xlim([xmin, xmax])
        ax.set_ylim([ymin, ymax])

        xmin_locator = dates.HourLocator(interval=1)
        xmaj_locator = dates.HourLocator(byhour=[0, 3, 6, 9, 12, 15, 18, 21, 24, 27])
        ax.xaxis.set_minor_locator(xmin_locator)
        ax.xaxis.set_major_locator(xmaj_locator)
        formatter = dates.DateFormatter('%H:%M')
        ax.xaxis.set_major_formatter(formatter)
        ax.set_xlabel("Time [hh:mm]")
        ax.xaxis_date()

        if not legend:
            ax.legend([])



import matplotlib.colors as mcolors
import numpy as np
from skimage import color, exposure

from movierender.render.pipelines._image_pipeline_base import ImagePipeline


class CompositeRGBImage(ImagePipeline):
    def _img(self, channel):
        r = self._renderer

        if type(self.zstack) is int:
            ix = r.image.ix_at(c=channel, z=self.zstack, t=r.frame)
            self.logger.debug(f"Retrieving frame {r.frame} of channel {channel} at z-stack={self.zstack} "
                              f"(index={ix})")
            imf = r.image.image(ix)
            if imf is not None:
                return r.image.image(ix).image
            return None
        elif type(self.zstack) is str:
            if self.zstack == "all-max":  # max projection
                self.logger.debug(f"Retrieving max z projection of frame {r.frame} and channel {channel}")
                return r.image.z_projection(frame=r.frame, channel=channel).image
            return None
        return None

    def __call__(self, *args, **kwargs):
        if 'channeldict' not in self._kwargs:
            raise Exception("Channel parameters needed to apply this pipeline.")
        channeldict = self._kwargs['channeldict']

        r = self._renderer

        dtype = None
        background = np.zeros((r.image.height, r.image.width) + (3,), dtype=np.float64)
        for name, settings in channeldict.items():
            channel = settings['id']
            _img = self._img(channel)
            if dtype is None:
                dtype = _img.dtype

            # Contrast enhancing by stretching the histogram
            if 'rescale' in settings and settings['rescale']:
                if type(settings['rescale']) is dict:
                    mini, maxi = settings['rescale']['range']
                    _img = exposure.rescale_intensity(_img, in_range=(mini, maxi))
                elif type(settings['rescale']) is bool and settings['rescale']:
                    _img = exposure.rescale_intensity(_img, in_range=tuple(np.percentile(_img, (0.1, 99.9))))

            _img = color.gray2rgb(_img)
            rgb_vector_color = mcolors.to_rgb(settings['color'])
            assert isinstance(rgb_vector_color, tuple)
            background += _img * rgb_vector_color * settings['intensity']

        return background.astype(dtype)



from .overlay import Overlay


class Text(Overlay):
    def __init__(self, text, **kwargs):
        assert text is not None, "Need text to render on axes."
        self.text = text
        super().__init__(**kwargs)

    def plot(self, ax=None, xy=None, fontdict=None, **kwargs):
        if ax is None:
            ax = self.ax
        assert ax is not None, "No axes found to plot overlay."
        if fontdict is None:
            fontdict = self._kwargs["fontdict"] if "fontdict" in self._kwargs else None
        if xy is None:
            xy = self._kwargs["xy"] if "xy" in self._kwargs else None

        x0, y0 = xy
        ax.text(x0, y0, self.text, **fontdict)



from ._experiment import Treatment
from ._histogram import ImageHistogram
from ._image import ScaleBar, Timestamp, secs_to_string
from ._roi import ImagejROI
from ._text import Text
from ._timeseries import DataTimeseries
from .overlay import Overlay
from .pixel_tools import PixelTools



import logging

import matplotlib.pyplot as plt
import numpy as np
import pandas as pd
import seaborn as sns
import skimage
from fileops.export.config import ConfigPanel
from fileops.image.ops import z_projection
from skimage.exposure import exposure

import movierender.overlays as ovl
from movierender.overlays import PixelTools

logger = logging.getLogger(__name__)


def plotimg(data, panel=None, **kwargs):
    imf = panel.image_file
    t = PixelTools(imf)

    ax = plt.gca()
    ax.cla()

    w_um, h_um = imf.width * imf.um_per_pix, imf.height * imf.um_per_pix
    sbar = ovl.ScaleBar(um=panel.scalebar, lw=3, xy=t.xy_ratio_to_um(0.05, 0.9), fontdict={'size': 9})
    hst = ovl.ImageHistogram(ax=ax, bins=50, color='white')

    if data["z"].unique().size > 1 and data["frame"].unique().size == 1 and data["channel"].unique().size == 1:
        _fr = data["frame"].iloc[0]
        _ch = data["channel"].iloc[0]
        img = z_projection(imf, _fr, _ch, projection='all-max')

        imgf = skimage.util.img_as_float(img.image)
        if f"channel-{_ch}" in panel.channel_render_parameters:
            ch_par = panel.channel_render_parameters[f"channel-{_ch}"]
            if "overlays" in ch_par and "histogram" in ch_par["overlays"]:
                # Overlay the histogram on the image plot
                hst.plot(img)
            if "color" in ch_par:
                imgf = exposure.rescale_intensity(imgf, in_range=tuple(np.percentile(imgf, (2, 99))), out_range=(0, 0.85))
                imgf = np.stack((imgf,) * 3, axis=-1) * panel.channel_render_parameters[f"channel-{_ch}"]["color"][1:4]

        ax.imshow(imgf, cmap='gray', extent=(.0, w_um, h_um, .0),
                  # origin='upper' if self.inv_y else 'lower',
                  origin='upper',
                  interpolation='none', aspect='equal',  # resample=False,
                  zorder=0)

        sbar.plot(ax=ax)
    ax.set_axis_off()


def render_static_montage(panel: ConfigPanel,
                          row=None, col=None):
    logger.debug("Making montage of image.")

    # create dataframe of images that will be plotted
    img_lst = [
        {
            'frame':   f,
            'channel': ch,
            'z':       z
        }
        for ch in panel.channels
        for z in panel.zstacks
        for f in panel.frames
    ]
    im_df = pd.DataFrame(img_lst)

    _s = 4.0
    g = sns.FacetGrid(im_df, row=row, col=col, height=_s)
    g = (g.map_dataframe(plotimg, panel=panel)
         .set_titles("{col_name}")
         .add_legend()
         )

    g.figure.tight_layout()
    plt.subplots_adjust(hspace=0, wspace=0.01, left=0, right=1, top=1, bottom=0)

    filepath = panel.configfile.parent / panel.filename
    g.savefig(filepath)
    return g



import os
import sys
from pathlib import Path

import typer
from typing_extensions import Annotated

from movierender.layouts import render_static_montage
from movierender.scripts._render_movie import render_movie

sys.path.append(Path(os.path.realpath(__file__)).parent.parent.parent.as_posix())

from fileops.export.config import read_config
from fileops.logger import get_logger, silence_loggers

log = get_logger(name='render-movie')


def render_configuration_file_cmd(
        cfg_path: Annotated[
            Path, typer.Argument(help="Name of the configuration file of the movie to be rendered")],
        show_file_info: Annotated[
            bool, typer.Argument(help="To show file metadata information before rendering the movie")] = True,
        overwrite_movie_file: Annotated[
            bool, typer.Option(help="Set true if you want to overwrite the file")] = False,
):
    if cfg_path.parent.name[0:3] == "bad":
        return

    log.info(f"Reading configuration file {cfg_path}")
    cfg = read_config(cfg_path)

    # make movies specified in configuration file
    for mov in cfg.movies:
        silence_loggers(loggers=[mov.image_file.__class__.__name__], output_log_file="silenced.log")
        if show_file_info:
            log.info(f"file {cfg_path}\r\n{mov.image_file.info.squeeze(axis=0)}")
        render_movie(mov, overwrite=overwrite_movie_file)

    # render panels specified in configuration file
    for pan in cfg.panels:
        silence_loggers(loggers=[pan.image_file.__class__.__name__], output_log_file="silenced.log")
        if show_file_info:
            log.info(f"file {cfg_path}\r\n{pan.image_file.info.squeeze(axis=0)}")
        render_static_montage(pan, row=pan.rows, col=pan.columns)



from __future__ import annotations

import asyncio
import io
import logging
import os
import pickle
from pathlib import Path
from typing import List, TYPE_CHECKING

import asyncpool
import imageio
import matplotlib.pyplot as plt
import moviepy.editor as mpy
import numpy as np
from fileops.export.config import ConfigMovie
from fileops.image import ImageFile
from fileops.image.exceptions import FrameNotFoundError
from fileops.pathutils import ensure_dir
from matplotlib.figure import Figure
# from memory_profiler import profile
from moviepy.video.io.bindings import mplfig_to_npimage

from movierender.render.pipelines import SingleImage, ImagePipeline

if TYPE_CHECKING:
    from movierender.overlays import Overlay


class ParallelMovieRenderer:
    layers: List[Overlay]
    image: ImageFile

    def __init__(self, fig: Figure, config: ConfigMovie, show_axis=False, invert_y=False, **kwargs):
        self._kwargs = {
            'fontdict': {'size': 10},
        }
        self._kwargs.update(**kwargs)

        self.fig = fig
        self.show_axis = show_axis
        self.ax = fig.gca()

        self.layers = []
        self.logger = logging.getLogger(__name__)

        self.time = 0
        self.frame = 0
        self.fps = config.fps
        self.duration = None
        self.bitrate = config.bitrate

        imf = config.image_file
        self._cfg = config
        self.image_pipeline: List[ImagePipeline] = []
        self.image = imf
        self.inv_y = invert_y
        self._last_f = imf.frames[-1]
        self._max_frame = max(self._cfg.frames)
        self._frame_offset = min(self._cfg.frames)  # used when frames start at a number greater than zero
        self._render = np.zeros((imf.width, imf.height), dtype=float)
        self._load_image()

        self._tmp = Path(os.curdir) / 'tmp' / 'render' / Path(imf.base_path).name
        ensure_dir(self._tmp)

    def __iter__(self):
        return self

    def __next__(self):
        if self.image_pipeline is not None:
            imp = self.image_pipeline
        else:
            imp = SingleImage(self)
        self.time = (self.time + 1) % self._max_frame

        return imp(invert_y=self.inv_y)

    def __getattr__(self, name):
        if name in self._kwargs:
            return self._kwargs[name]
        else:
            raise AttributeError("No such attribute: " + name)

    def __deepcopy__(self):
        # check there is only one figure to render
        figures = {self.fig}
        for imgp in self.image_pipeline:
            if imgp.ax is not None:
                figures.add(imgp.ax.get_figure())
        assert len(
            figures) == 1, "There are more than one figure in this MovieRenderer object and thus cannot be deep-copied."

        # copy new figure and axes from initial instance setup
        buf = io.BytesIO()
        pickle.dump(self.fig, buf)
        buf.seek(0)
        new_fig = pickle.load(buf)

        # extract equivalences between axes
        newax = {old_ax: new_ax for old_ax, new_ax in zip(self.fig.axes, new_fig.axes)}
        # create new MovieRenderer instance
        new_mvr = MovieRenderer(new_fig, self.image, fps=self.fps, bitrate=self.bitrate, show_axis=self.show_axis)
        # copy all image pipelines
        for imgp in self.image_pipeline:
            c_imgp = imgp.__class__
            new_ax = newax[imgp.ax] if imgp.ax is not None else None
            new_imgp = c_imgp(ax=new_ax)
            for d in imgp.__dict__:
                if d not in ['uuid', 'ax', '_renderer']:
                    new_imgp.__dict__.update({d: imgp.__dict__[d]})
            new_imgp.__dict__.update({'_renderer': new_mvr})

            new_mvr.image_pipeline.append(new_imgp)

        # copy all layer render
        for layer in self.layers:
            lcls = layer.__class__
            new_ax = newax[imgp.ax] if imgp.ax is not None else None
            nl = lcls(ax=new_ax)
            for d in layer.__dict__:
                if d not in ['uuid', 'ax', '_renderer']:
                    nl.__dict__.update({d: layer.__dict__[d]})
            nl.__dict__.update({'_renderer': new_mvr})

            new_mvr.layers.append(nl)

        return new_mvr

    def _mem_clean(self):
        self._renderer = None
        self.image_pipeline = None
        self.fig = None
        self.ax = None
        self.layers = None
        self.logger = None
        self.image_pipeline = None
        self.image = None
        self._last_f = None
        self._render = None
        self._tmp = None

    def _load_image(self):
        assert len(self.image.frames) > 1, "More than one frame needed to make a movie."
        self.logger.info(f"Loaded {self.image.image_path}. WxH({self.image.width},{self.image.height}), "
                         f"channels: {len(self.image.channels)}, "
                         f"frames: {len(self.image.frames)}, stacks: {len(self.image.zstacks)}")
        self.duration = len(self.image.frames) / self.fps
        self.logger.info(f"Total duration {self.duration:.3f}[s]")

        self._kwargs.update({
            'pix_per_um': self.image.pix_per_um,
            'timestamps': self.image.timestamps,
            'dt':         self.image.time_interval,
            'n_frames':   len(self.image.frames),
            'n_channels': len(self.image.channels),
            'n_zstacks':  len(self.image.zstacks),
            'width':      self.image.width,
            'height':     self.image.height,
            # 'ranges':     self.image.intensity_ranges
        })

    def render(self, filename=None):
        """
        Render frames of a movie in parallel.
        """

        def make_frame_mpl(t):
            self.time = t
            # calculate frame given time
            _fr_from_t = int(round(self.fps * t)) + self._frame_offset
            self.frame = min(_fr_from_t, self._max_frame)

            if self.frame == self._last_f:
                return self._render

            self._last_f = self.frame
            self._render = imageio.imread(self._tmp.joinpath(f"f{self.frame:05d}.png"))
            # self.logger.debug(f"loaded image of shape {self._render.shape}")
            return self._render

        # @profile
        async def render_frame(frame, rq):
            img_path = self._tmp.joinpath(f"f{frame:05d}.png")
            if os.path.exists(img_path):
                self.logger.warning(f"skipping frame {frame}.")
                return

            # make a copy of this instance so it can run in parallel
            mvr = self.__deepcopy__()

            mvr.frame = frame

            # clear axes of all objects
            mvr.ax.cla()
            for ovrl in mvr.layers:
                if ovrl.ax is not None:
                    ovrl.ax.cla()
            for imgp in mvr.image_pipeline:
                if imgp.ax is not None:
                    imgp.ax.cla()
                if not mvr.show_axis and imgp.ax is not None:
                    imgp.ax.set_xticklabels([])
                    imgp.ax.set_yticklabels([])
                    imgp.ax.set_xticks([])
                    imgp.ax.set_yticks([])

            for imgp in mvr.image_pipeline:
                try:
                    ppu = mvr.pix_per_um if mvr.pix_per_um is not None else 1
                    ext = [0, mvr.width / ppu, 0, mvr.height / ppu]
                    ax = imgp.ax if imgp.ax is not None else mvr.ax
                    ax.imshow(imgp(), cmap='gray', extent=ext,
                              origin='upper' if self.inv_y else 'lower',
                              interpolation='none', aspect='equal',
                              zorder=0)
                except FrameNotFoundError:
                    continue

            for ovrl in mvr.layers:
                kwargs = mvr._kwargs.copy()
                kwargs.update(**ovrl._kwargs, show_axis=mvr.show_axis)
                ovrl.plot(ax=mvr.ax if ovrl.ax is None else None, **kwargs)

            for ovrl in mvr.layers:
                if not ovrl.show_axis and ovrl.ax is not None:
                    ovrl.ax.set_xticklabels([])
                    ovrl.ax.set_yticklabels([])
                    ovrl.ax.set_xticks([])
                    ovrl.ax.set_yticks([])
            mvr.fig.tight_layout()

            img = mplfig_to_npimage(mvr.fig)  # RGB image of the figure
            imageio.imwrite(img_path, img)

            plt.close(mvr.fig)
            mvr._mem_clean()
            del mvr

            await rq.put(f"Frame {frame} saved to {img_path}.")

        async def result_reader(queue):
            while True:
                value = await queue.get()
                if value is None:
                    break
                self.logger.debug("Got value! -> {}".format(value))

        async def run():
            result_queue = asyncio.Queue()
            reader_future = asyncio.ensure_future(result_reader(result_queue), loop=loop)
            num_procs = os.cpu_count() * 2

            # Start a worker pool with num_procs coroutines, invokes `render_frame` and waits for it to complete.
            async with asyncpool.AsyncPool(loop, num_workers=num_procs,
                                           name="RenderPool",
                                           logger=logging.getLogger("RenderPool"),
                                           worker_co=render_frame,
                                           log_every_n=10) as pool:
                for i in self.image.frames:
                    await pool.push(i, result_queue)

            await result_queue.put(None)
            await reader_future

        # Start of video render method
        if filename is None:
            _, filename = os.path.split(self._file)
            filename += ".mp4"

        loop = asyncio.get_event_loop()
        loop.run_until_complete(run())

        animation = mpy.VideoClip(make_frame_mpl, duration=self.duration)
        animation.write_videofile(filename, fps=self.fps, bitrate=self.bitrate)
        animation.close()

    def __repr__(self):
        return f"<MovieRender object at {hex(id(self))}> with {len(self._kwargs)} arguments."



import os
from pathlib import Path

from fileops.export.config import ConfigMovie
from fileops.logger import get_logger

from movierender.render import MovieRenderer


class BaseLayoutComposer:
    log = get_logger(name='BaseLayoutComposer')

    def __init__(self, movie: ConfigMovie,
                 prefix='', suffix='',
                 overwrite=False,
                 **kwargs):
        self._overwrite = overwrite
        self._movie_configuration_params = movie
        self.renderer: MovieRenderer | None = None
        self.dpi = 326

        self.fig_title = movie.title
        self.ax_lst = list()

        im = movie.image_file
        fname = movie.movie_filename if len(movie.movie_filename) > 0 else im.image_path.name
        self.filename = prefix + fname
        if len(suffix) > 0:
            self.filename += "." + suffix
        self.filename += ".mp4"
        self.base_folder = movie.configfile.parent
        self.save_file_path = Path(self.base_folder) / self.filename

        if os.path.exists(self.save_file_path):
            if os.path.getsize(self.save_file_path) < 300:  # if size is too small, treat it as if the file didn't exist
                overwrite = True
            if not overwrite:
                self.log.warning(f'File {self.filename} already exists in folder {self.base_folder}.')
                raise FileExistsError

    def make_layout(self):
        raise NotImplementedError

    def render(self):
        self.log.info(f"Rendering movie into file {self.save_file_path}.")
        if self.renderer is None:
            raise AttributeError("Need to call method make_layout before trying to render.")
        self.renderer.render(filename=str(self.save_file_path), test=False)



from movierender.render import MovieRenderer


class Overlay(object):
    def __init__(self, ax=None,  **kwargs):
        self.layers = [self]
        self._kwargs = kwargs
        self._renderer: MovieRenderer | None = None
        self.ax = ax

    def __radd__(self, ovrl):
        if isinstance(ovrl, MovieRenderer):
            ovrl.layers += self.layers
            self._renderer = ovrl
            self.show_axis = self._renderer.show_axis
            return ovrl

        self.layers.append(ovrl)
        return self

    def plot(self, ax, **kwargs):
        pass



import os
from pathlib import Path

import numpy as np
from fileops.export.config import ConfigMovie
from fileops.logger import get_logger
from matplotlib import pyplot as plt, gridspec

import movierender.overlays as ovl
from movierender import MovieRenderer, SingleImage, CompositeRGBImage
from movierender.overlays.pixel_tools import PixelTools

log = get_logger(name='movielayout')

green = [0, 1, 0]


def make_movie(movie: ConfigMovie, prefix='', suffix='', folder='.', overwrite=False):
    im = movie.image_file
    assert len(im.channels) >= 2, 'Image series contains less than two channels.'
    print(im.info)
    print(im.zstacks)
    filename = prefix + os.path.basename(im.image_path) + suffix + ".all_z.mp4"
    base_folder = os.path.abspath(folder)
    path = os.path.join(base_folder, filename)
    if os.path.exists(path):
        if not overwrite:
            log.warning(f'File {filename} already exists in folder {base_folder}.')
            return

    Path(path).touch()
    log.info(f'Making movie {filename} from file {os.path.basename(im.image_path)} in folder {base_folder}.')
    t = PixelTools(im)

    n_cols = 4
    n_rows = int(np.ceil(len(im.zstacks) / n_cols))
    ar = float(im.height) / float(im.width)
    log.info(f'aspect ratio={ar}.')
    log.info(f'number of stacks={len(im.zstacks)}, n_rows={n_rows}.')

    fig = plt.figure(frameon=False, figsize=(10, 10), dpi=150)
    gs = gridspec.GridSpec(nrows=n_rows, ncols=n_cols)

    # set positions of scale bars and text
    scale_xy = t.xy_ratio_to_um(0.80, 0.05)
    ztext_xy = t.xy_ratio_to_um(0.70, 0.95)

    # first axes is for BF
    ax_ch1 = fig.add_subplot(gs[0, 0])
    fig.subplots_adjust(left=0.125, right=0.9, bottom=0.1, top=0.99, wspace=0.01, hspace=0.01)
    movren = MovieRenderer(fig=fig,
                           config=movie,
                           fontdict={'size': 12}) + \
             ovl.ScaleBar(um=movie.scalebar, lw=3, xy=scale_xy, fontdict={'size': 7}, ax=ax_ch1) + \
             ovl.Timestamp(xy=t.xy_ratio_to_um(0.02, 0.95), va='center', ax=ax_ch1) + \
             SingleImage(ax=ax_ch1, channel=0)

    # rest of grid is for the channel of interest across all Z
    for grid_pos in zip(np.repeat(np.arange(n_rows), n_cols), np.tile(np.arange(n_cols), n_rows)):
        row, col = grid_pos
        z_ix = row * n_cols + col - 1
        if grid_pos != (0, 0) and z_ix < len(im.zstacks):
            ax_z = fig.add_subplot(gs[grid_pos[0], grid_pos[1]])
            zpos = im.zstacks[z_ix]
            movren = movren + \
                     CompositeRGBImage(ax=ax_z, zstack=zpos, channeldict={
                         'Channel2': {
                             'id':        1,
                             'color':     green,
                             'rescale':   True,
                             'intensity': 1.0
                         },
                     }) + \
                     ovl.ScaleBar(um=movie.scalebar, lw=3, xy=scale_xy, fontdict={'size': 7}, ax=ax_z) + \
                     ovl.Text(f'z={zpos * im.um_per_z} [um]', xy=ztext_xy, fontdict={'size': 7}, ax=ax_z)
    movren.render(filename=path, test=False)



import numpy as np
from fileops.image.imagemeta import MetadataImage
from skimage import exposure

from movierender.render.pipelines._image_pipeline_base import ImagePipeline


class SingleImage(ImagePipeline):
    def __call__(self, *args, channel=0, adjust_exposure=True, **kwargs):
        r = self._renderer

        ix = r.image.ix_at(c=channel, z=self.zstack, t=r.frame)
        self.logger.debug(f"Retrieving frame {r.frame} of channel {channel} at z-stack={self.zstack} "
                          f"(index={ix})")
        mimg = r.image.image(ix)
        img = mimg.image if (type(mimg) is MetadataImage and mimg.image is not None) else np.zeros((r.width, r.height))
        if adjust_exposure:
            p2, p98 = np.percentile(img, (2, 98))
            img = exposure.rescale_intensity(img, in_range=(p2, p98))
        return img



import copy
import math

from fileops.export.config import ConfigMovie
from fileops.logger import get_logger
from matplotlib import pyplot as plt, gridspec

import movierender.overlays as ovl
from movierender import MovieRenderer, CompositeRGBImage
from movierender.overlays.pixel_tools import PixelTools
from ._base_composer import BaseLayoutComposer


class LayoutColumnComposer(BaseLayoutComposer):
    log = get_logger(name='LayoutColumnComposer')

    def __init__(self,
                 movie: ConfigMovie,
                 columns: int = 2,
                 **kwargs):
        super().__init__(movie, **kwargs)

        self.n_columns = columns

    def make_layout(self):
        movie = self._movie_configuration_params
        t = PixelTools(movie.image_file)

        fig = plt.figure(figsize=(5 * self.n_columns, 5.5), dpi=self.dpi)
        fig.suptitle(self.fig_title)

        if len(movie.channels) > 1:
            n_channels = len(movie.channels)
            rows = math.ceil(n_channels / self.n_columns)
            gs = gridspec.GridSpec(nrows=rows, ncols=self.n_columns)
            self.log.debug(f"making frid of {rows} rows and {self.n_columns} columns.")

            for i in range(n_channels):
                self.ax_lst.append(fig.add_subplot(gs[0, i]))
            fig.subplots_adjust(left=0.125, right=0.9, bottom=0.1, top=0.99, wspace=0.01, hspace=0.01)
        else:
            self.ax_lst.append(fig.gca())

        self.renderer = MovieRenderer(fig=fig,
                                      config=movie,
                                      fontdict={'size': 12})

        for ax, ch_cfg_ix in zip(self.ax_lst, movie.channel_render_parameters):
            ch_cfg = movie.channel_render_parameters[ch_cfg_ix]
            self.renderer += ovl.ScaleBar(um=movie.scalebar, lw=3,
                                          xy=t.xy_ratio_to_um(0.80, 0.05),
                                          fontdict={'size': 9},
                                          ax=ax)
            self.renderer += ovl.Timestamp(xy=t.xy_ratio_to_um(0.02, 0.95), va='center', ax=ax)
            self.renderer += CompositeRGBImage(ax=ax,
                                               zstack=movie.zstack_fn,
                                               channeldict={
                                                   ch_cfg['name']: {
                                                       'id':        ch_cfg_ix,
                                                       'color':     ch_cfg['color'][1:] if (
                                                               isinstance(ch_cfg['color'], tuple) and
                                                               len(ch_cfg['color']) > 3
                                                       ) else ch_cfg['color'],
                                                       'rescale':   True,
                                                       'intensity': 1.0
                                                   },
                                               })
            self.renderer += ovl.Text(f'{ch_cfg["name"]}',
                                      xy=t.xy_ratio_to_um(0.70, 0.95),
                                      fontdict={'size': 7, 'color': 'white'}, ax=ax)

    def add_overlay(self, overlay: ovl.Overlay):
        for ax in self.ax_lst:
            _o = copy.deepcopy(overlay)
            _o.ax = ax
            self.renderer += _o



from fileops.export.config import ConfigMovie
from fileops.logger import get_logger
from matplotlib import pyplot as plt

import movierender.overlays as ovl
from movierender import MovieRenderer, CompositeRGBImage
from movierender.overlays.pixel_tools import PixelTools
from ._base_composer import BaseLayoutComposer


class LayoutCompositeComposer(BaseLayoutComposer):
    log = get_logger(name='LayoutColumnComposer')

    def __init__(self,
                 movie: ConfigMovie,
                 **kwargs):
        super().__init__(movie, **kwargs)

    def make_layout(self):
        movie = self._movie_configuration_params
        t = PixelTools(movie.image_file)

        fig = plt.figure(figsize=(5, 5.5), dpi=self.dpi)
        fig.suptitle(self.fig_title)

        # only one axes is rendered
        ax = fig.gca()
        self.ax_lst.append(ax)

        self.renderer = MovieRenderer(fig=fig,
                                      config=movie,
                                      fontdict={'size': 12})

        self.renderer += ovl.ScaleBar(um=movie.scalebar, lw=3,
                                      xy=t.xy_ratio_to_um(0.80, 0.05),
                                      fontdict={'size': 9},
                                      ax=ax)
        self.renderer += ovl.Timestamp(xy=t.xy_ratio_to_um(0.02, 0.95), va='center', ax=ax)
        self.renderer += CompositeRGBImage(ax=ax,
                                           zstack=movie.zstack_fn,
                                           channeldict={
                                               ch_cfg['name']: {
                                                   'id':        cix,
                                                   'color':     ch_cfg['color'][1:] if (
                                                           isinstance(ch_cfg['color'], tuple) and
                                                           len(ch_cfg['color']) > 3
                                                   ) else ch_cfg['color'],
                                                   'rescale':   True,
                                                   'intensity': 1.0
                                               } for cix, ch_cfg in movie.channel_render_parameters.items()})



from unittest import TestCase

from typer.testing import CliRunner

from movierender.scripts.render import app


class TestRender(TestCase):
    def __init__(self, *args):
        super().__init__(*args)
        self.runner = CliRunner()

    def test_render_movie_different_frames(self):
        command_name = "movie"

        args = [command_name, "example_data/test_frames_movie.cfg"]
        result = self.runner.invoke(app, args)
        self.assertEqual(result.exit_code, 0)

    def test_render_movie_swap_channels(self):
        command_name = "movie"

        args = [command_name, "example_data/test_frames_movie_ch_swap.cfg"]
        result = self.runner.invoke(app, args)
        self.assertEqual(result.exit_code, 0)

    def test_render_panel(self):
        command_name = "panel"

        args = [command_name, "example_data/test_panels.cfg"]
        result = self.runner.invoke(app, args)
        self.assertEqual(result.exit_code, 0)

    def test_render_folder(self):
        command_name = "folder"

        args = [command_name, "example_data"]
        result = self.runner.invoke(app, args)
        self.assertEqual(result.exit_code, 0)


if __name__ == "__main__":
    app()



import os
import sys
from pathlib import Path

import typer
from typing_extensions import Annotated

from movierender.layouts import LayoutColumnComposer, LayoutCompositeComposer

sys.path.append(Path(os.path.realpath(__file__)).parent.parent.parent.as_posix())

from fileops.export.config import read_config, ConfigMovie
from fileops.logger import get_logger, silence_loggers

log = get_logger(name='render-movie')


def render_movie(mov: ConfigMovie, overwrite=False):
    if len(mov.image_file.frames) == 1:
        log.warning("only one frame, skipping static image")
        return
    elif len(mov.image_file.frames) > 1:
        mv_kwargs = dict(overwrite=overwrite)
        # what follows is a list of supported layouts
        if mov.layout in ["twoch", "two-ch"]:
            lytcomposer = LayoutColumnComposer(mov, columns=2, **mv_kwargs)
        elif mov.layout == "twoch-comp":
            lytcomposer = LayoutCompositeComposer(mov, **mv_kwargs)
        else:
            raise ValueError(f"No supported layout in the rendering of {mov.movie_filename}.")

        lytcomposer.make_layout()
        lytcomposer.render()


def render_movie_cmd(
        cfg_path: Annotated[
            Path, typer.Argument(help="Name of the configuration file of the movie to be rendered")],
        show_file_info: Annotated[
            bool, typer.Argument(help="To show file metadata information before rendering the movie")] = True,
        overwrite_movie_file: Annotated[
            bool, typer.Option(help="Set true if you want to overwrite the file")] = False,
):
    if cfg_path.parent.name[0:3] == "bad":
        return
    log.info(f"Reading configuration file {cfg_path}")
    cfg = read_config(cfg_path)

    # make movies specified in configuration file
    for mov in cfg.movies:
        silence_loggers(loggers=[mov.image_file.__class__.__name__], output_log_file="silenced.log")
        if show_file_info:
            log.info(f"file {cfg_path}\r\n{nfo.squeeze(axis=0)}")
        render_movie(mov, overwrite=overwrite_movie_file)



from fileops.image import ImageFile


class PixelTools:
    def __init__(self, cimg: ImageFile):
        self.width = 0
        self.height = 0
        self.um_per_pix = 0
        for d in cimg.__dict__:
            self.__dict__.update({d: cimg.__dict__[d]})

    def xy_ratio_to_pixels(self, x, y):
        # nfo = self.info
        assert 0 <= x <= 1, "x is not in expected range."
        assert 0 <= y <= 1, "y is not in expected range."
        return x * self.width, y * self.height

    def xy_ratio_to_um(self, x, y):
        assert 0 <= x <= 1, "x is not in expected range."
        assert 0 <= y <= 1, "y is not in expected range."
        return x * self.width * self.um_per_pix, y * self.height * self.um_per_pix



import numpy as np
from fileops.image import MetadataImage

from .overlay import Overlay


class ImageHistogram(Overlay):
    def plot(self, mdi: MetadataImage, bins=None, ax=None, color=None, **kwargs):
        if ax is None:
            ax = self.ax
        assert ax is not None, "No axes found to plot overlay."

        bins = bins if bins is not None else self._kwargs.get("bins", 10)
        color = color if color is not None else self._kwargs.get("color", "magenta")

        axi = ax.inset_axes(
            (0.5, 0.5, 0.47, 0.47),
            facecolor=color,
            frameon=False
        )

        hist, bins = np.histogram(mdi.image.ravel(), bins=bins)
        axi.hist(bins[:-1], bins, weights=hist, histtype='step', color=color, zorder=10)
        # axi.set_xlabel('intensity', color='white')
        # axi.set_ylabel('pixel count', color='white')
        axi.tick_params(axis='both', colors=color)
